
/* a little mix and match of eventqueue (untested)
 * and tachin (rpm) measurements 
 */

#include "heap.h"
#include "eventqueue.h"

#include <inttypes.h>
#include <avr/signal.h>
#include <avr/io.h>
#include <avr/interrupt.h>


struct tachin_t tachin;
struct pwdisplay_t pwdisplay;
struct heap_t *ho, *hn; /* pointers for heap_overflow and heap_normal */

uint16_t now;

void tachin_reset(void)
{
	int8_t	i;	// reset TACH structures
	for(i=sizeof(struct tachin_t)-1; i>=0; i--){
		((char*)(&tachin))[i] = 0;
	}
}
uint8_t pw2comm(uint8_t i)
{
	return 	((char*)(&pwdisplay))[i];
}

/* Why doing the eventqueue with timer2 (8 bit) when timer3 has 16 bit
 * precision?
 * If we use timer2, we have 2 (or even 3, since timer2's pwm output is shared!)
 * more PWM pins. This is the main reason, it really matters.
 * On the avr writing the 16 bit registers involves some magic.
 * The 8 bit can be handled a little safer, less risk of losing a compare match..
 * The cost of having an interrupt at least every 255 * 4 usec, thus 
 * about 1000/sec is really low (even compared to the 10000Hz interrupt before
 */

void eventqueue_init(void)
{
	int8_t	i;	// reset TACH structures
	TCCR2 = _BV(CS21) | _BV(CS20);	/* set prescaler to division 64 (4usec) */

#ifdef PWANAL
	TCCR3B = _BV(CS31) | _BV(CS30); /* set prescaler to division 64 (4usec) page 133 */ 
#endif
	
	/* set mode (up, not_oneshot, ...) NOTHING to do, using default */

	/* rising edge */

	/* enable interrupt */
	TIMSK |= _BV(OCIE2);
//	ETIMSK |= _BV(OCIE3A); 

	/* reset struct */
	tachin_reset();
	for(i=sizeof(struct pwdisplay_t); i>=0; i--){
		((char*)(&pwdisplay))[i] = 0;
	}
	 heap_init(&h1); 
	 ho = &h1;
	 heap_init(&h2); 
	 hn = &h2;
}
void pwdisplay_from_tachin(uint16_t now)
{
	/* 1 pulsewidth is usually not more than 8msec = 2000, so 32 will fit */
	pwdisplay.pwsum = tachin.pwsum >> ROTBY;
	pwdisplay.persum = tachin.persum >> ROTBY ;
	pwdisplay.bad = tachin.edge_bad;

/*	pwdisplay.pwsqrsum = tachin.pwsqrsum;
	pwdisplay.persqrsum = tachin.persqrsum;
 */
	pwdisplay.dirty = 1; // signal that it has not been displayed
}

/* Called from dispatcher:
 * inserts an event and destroys the first in struct heap hn 
 */
__inline__ void event_insert_destroy(key_t k, val_t v)
{
	if(k >= now) { // we need the heap_normal(hn)
		heap_replace1st(hn, k, v);
	} else { // thanks to Michael for pointing this out:
		heap_insert(ho, k, v); 
		heap_del1st(hn);
	} // k == now could be splitted (but it should never happen, right?)
}
/* Called from dispatcher:
 * inserts an event and does NOT destroy the first in struct heap hn 
 * Note: in dispatcher we want to destroy hn[HEAPROOT] just once. For any 
 * further events to be inserted, this is the function!
 */
__inline__ void event_insert(key_t k, val_t v)
{
	if(k >= now) { // we need the heap_normal(hn)
		heap_insert(hn, k, v);
	} else { 
		heap_insert(ho, k, v); 
	} // k == now could be splitted (but it should never happen, right?)
}

/* Keep val == 0 reserved, real events must have nonzero val.
 * The dispatcher is responsible for destruction of current event.
 * Think of inserting the next cyl's event when this is just accomplished.
 */
__inline__ void dispatcher(val_t v)
{
	// static uint8_t ledcnt3;
	// do what you want according to v, 
	// we are always doing hn[HEAPROOT] at this point

	// eg we just started cyl1_ignition, we designate 2 events:
	//event_insert_destroy(time_of_cyl1_ignstop, val_cyl1_ignstop);
	//event_insert(time_of_cyl3_ignstart, val_cyl3_ignstart);
}

/*
SIGNAL (SIG_OUTPUT_COMPARE3) {
static uint16_t tmr_before;
	uint16_t next;

	now = TCNT3;	// read counter

	while((heaplast >= HEAPROOT) && (heapkey(HEAPROOT) == now)){
		dispatcher(heapval(HEAPROOT));
		heap_del1st();
		heap_insert(now - 1, 0);
	}
	if (heaplast < HEAPROOT) OCR3A = now - 1 ;// the furthest away timepoint
	else OCR3A = heapkey(HEAPROOT);
}
*/

__inline__ uint16_t readTMR2(void)
{
	static uint8_t upper, prev_tcnt2;
	struct heap_t *h;
	uint8_t t;

	t=TCNT2;
	if (t < prev_tcnt2){	// an overflow occured
		if(++upper == 0){ // overflow, 
			while(hn->heaplast >= HEAPROOT){ // we consume the old heap (rarely needed)
				dispatcher(heapval(hn, HEAPROOT));
			}; // hn is now empty
			h = ho;	/* we need to swap the 2 heaps */
			ho = hn;
			hn = h;
		}
	}
	prev_tcnt2 = t;
	return upper << 8 | t;
}
/* only slightly different.
 * if we don't have frequent events, we can set OCR2
 * to now+255 (previous OCR2-1) or leave OCR2 alone (-> now+256).
 * In the latter case we'll need to begin the interrupt handler with this function:
 * at the beginning of COMPARE2 interrupt,
 * Assumes that NEVER executed twice in the same 4 usec (it would increase the upper bits)
 * 
 */
__inline__ uint16_t readTMR2_incifsame(void)
{
	static uint8_t upper, prev_tcnt2;
	struct heap_t *h;
	uint8_t t;

	t=TCNT2;
	if (t <= prev_tcnt2){	// an overflow occured EVEN if the value is SAME as before
		if(++upper == 0){ // overflow, we need to swap the 2 heaps
			while(hn->heaplast >= HEAPROOT){ // we consume the old heap (rarely needed)
				dispatcher(heapval(hn, HEAPROOT));
			}; // hn is now empty
			h = ho;	
			ho = hn;
			hn = h;
		}
	}
	prev_tcnt2 = t;
	return upper << 8 | t;
}
/* for 8 bit timer it take a little more:
 if event is too far, we set to 255 distance */
void setalarm2(key_t k)
{
	uint16_t now;
	now = readTMR2();
	if(k - now > 255) OCR2 = (0xFF & now) - 1;	// TODO: check overflow
	else OCR2 = k & 0xFF;
}

/* we always disable interrupts since the heap is manipulated from interrupt
 * routine. A conflict would result in trashed structures.
 * after all the events due now are served 
 * we insert the next event
 *
 * It seems that the action needed at an event_insert from userspace is
 * almost exactly the same as when an OCR compare match interrupt was fired.
 * Can we call the interrupt handler (after cli()?
 * so that reti will return an stack aligned right?
 */

SIGNAL (SIG_OUTPUT_COMPARE2) {
	uint16_t now, nextev;

	/* the interrupts are disabled. Only we are working on the 
	* heap at the moment
	*/

	now = readTMR2();	// read counter

	do{
	while((hn->heaplast >= HEAPROOT) && (heapkey(hn, HEAPROOT) == now)){
		dispatcher(heapval(hn, HEAPROOT));
	}
	if (hn->heaplast < HEAPROOT || heapkey(hn,HEAPROOT)-now > 255) OCR2 = (now & 0xFF) - 1 ;// the furthest away timepoint if no events, or too much away in the future
	else OCR2 = heapkey(hn, HEAPROOT) & 0xFF ;
	/* note, that OCR2 could theoretically be set to (now & 0xff)
	 * instead of (now & 0xFF) - 1 since compare match is
	 * blocked for the next TNCT2 count!
	 * However then readTMR2_incifsame should be used at the beginning of interrupt-routine
	 * which requires more care. The difference in efficiency is minimal.
	 * This OCR2 setting could be moved to the end of the interrupt handler, but
	 * we would need code duplication to avoid missing a TCNT2 value .
	 */

	/* Assuming we did everything we needed at the moment,
	 * clear interrupt flag (maybe OCR matched in the meantime.
	 * Unfortunately there is a race condition between checking the
	 * TCNT and flag-clear. Recheck TCNT2. probable involves an outer cycle /eg do-while/
	 */
	 TIFR=_BV(OCF2); /*  TIFR |= _BV(OCF2); could clear OTHER flags! */
	now = readTMR2(); 

	} while ((hn->heaplast >= HEAPROOT) && (heapkey(hn, HEAPROOT) == now )); // go back if there is an event very close.. Should we change it to (heapkey(hn, HEAPROOT) - now < 2  ??
	// the old condition (TCNT2 counted) could result in infinite loop if even the short path takes too long to execute!
}
