 /***********************************************************
  *   Program:    LCD driver
  *   Author:     Marcell Gal <cell@x-dsl.hu>
  *   Comments:   HITACHI LCD driver
  ************************************************************/
#include "global.h"
#include "lcd.h"
#include "comm.h"
#include "eventqueue.h"


#include <inttypes.h>
#include <avr/pgmspace.h>
#include <avr/delay.h>
#ifdef LCD
/*
  TODO
    The wait parameter should be checked. Currently a wait==0 means
    the busyflag is polled
*/


/* we keep an in memory copy of the LCD data, then update one at a time */
char 	lcd_cache[LCD_CHARS];
uint8_t	dirty[LCD_CHARS/8];

/* to avoid lotsof lcd_goto commands we try to keep track of where the cursor is.
 * NOTE: This is index in lcd_cache, not address in CGRAM  */
lcdpos_t lcd_went_to;

/* what should be checked next: */
lcdpos_t lcd_pos_to_peek;

/* position in cache that will be written next time by eg. printhex */
lcdpos_t lcd_next_pos;


uint8_t lcd_cache2comm(uint8_t i)
{
//	if(i == 0) return 'L';
//	if(i == 1) return '=';
	return lcd_cache[i];
}

/* this uses lcd_busy() to check if the lcd is ready to accept a new command */
uint8_t can_we_go(void)
{
  return !lcd_busy();
}

void busywait(void) {
  while (lcd_busy()) 
    ;//    lcd_delay(10);
}

/* when new data is available this should be called 
 * This writes interesting variables to the LCD */
void lcd_newdata(void)
{

}

/* the char at index should be sent to the LCD */
__inline__ void dirtyat(lcdpos_t index)
{
	dirty[index>>3] |= 1 << (index & 7) ;
}

/* the char at index is synchronized with LCD thus sent. - no need to send again -
 * return old flag (Test and Clear, effectively) */
__inline__ uint8_t TaC(lcdpos_t index)
{
	uint8_t t, ret, mask;
	t = index>>3;
	mask = 1 << (index & 7) ;
	ret = dirty[t] & mask;
	dirty[t] &= ~mask;
	return ret;
}

/* checks if the char at index is dirty */
__inline__ char isdirty(lcdpos_t index)
{
	return dirty[index>>3] & (1 << (index & 7)) ;
}

/* lcd_refresh should be called frequently, eg. 32..100 times/sec 
 * LCD is completely updated after  appr. LCD_CHARS invocations
 */
void lcd_refresh(void)
{
#ifdef COMM_AUTO_DUMP
        static int lowerfreq; // only to limit comm.c traffic
#endif

	if(!can_we_go()) return;

#if LCD_CHARS == 32
	lcd_pos_to_peek &= (LCD_CHARS - 1);	/* handle overflow */
#else
	if (lcd_pos_to_peek > (LCD_CHARS - 1)) lcd_pos_to_peek = 0;
#endif
	if(isdirty(lcd_pos_to_peek)){ /* it was dirty, we have to send char */
		if(lcd_went_to == lcd_pos_to_peek){
		  //MIK TODO: shouldn't lcd_went_to be updated at a regular write to lcd?
		  //Cell: yes, since autoincreas is turned on! 
			lcd_write(lcd_cache[lcd_pos_to_peek]);
			TaC(lcd_pos_to_peek);
			lcd_pos_to_peek++;
		} else { /* cursor was not there */
			lcd_go2index(lcd_pos_to_peek);
			/* we will lcd_putch at next invocation */
		}
		setleds(); // port was screwed, let's set it back
	} else { /* it was clear */
		lcd_pos_to_peek++;
	};

#ifdef COMM_AUTO_DUMP
	if(--lowerfreq <= 0 ){
		char i;
		pushfunc(&lcd_cache2comm, LCD_CHARS);
		pushfunc(&pw2comm,  16 );

		lowerfreq = 30000 ;
		for (i=LCD_CHARS/8; i>=0; i--){	// set lcd_cache dirty to ensure refresh
			dirty[(uint8_t)i] = 0xFF;
			lcd_went_to = 255;
		}
	}
#endif
}

void update_char(char c, lcdpos_t index)
{
	// TODO: Is this correct?
	// yes, but not strictly necessary. 
	// this will make refresh slower on the average:
	// When writing 5 chars (eg. a number), lcd_next_pos will point to the last
	// position. Solution: we can normally go lcd_next_pos-- direction; MIK? ;
	lcd_next_pos = index + 1;
	if (lcd_next_pos > (LCD_CHARS - 1))
	  lcd_next_pos=0;

	if(lcd_cache[index] == c) return; /* nothing to do */
	lcd_cache[index] = c;
	dirtyat(index);

}

/* this gets a string */
void update_chars(char *c, lcdpos_t index)
{
	lcdpos_t i, z;
	z=0;
	for(i=index; i<LCD_CHARS && c[z]; i++){
		update_char(c[z++], i);
	}
}

/* puts a sign character to index, and returns unsigned value */
uint16_t stripsign(int bin, lcdpos_t index)
{
	if(bin<0){
		update_char('-', index);
		return 0-bin;
	}
	update_char('+', index);	/* or ' ' */
	return bin;
}

/* FAST conversion!
 * result is exactly 5 LCD chars updated starting from index ...
 * no sign, see stripsign() if you need it
 */
uint16_t BCDbits[] = {80,800,8000};
void int2BCD5(uint16_t bin, lcdpos_t index)
{
  int8_t i;
  uint8_t c, tenpower8;
  uint16_t result, t;

  tenpower8 = 2;

  t=40000; c=4;
  result = '0';	/* 1 digit of the number */
  for (i=14; i>=0; i--) {  // we don't start at 15, and we also want the last (0) element
    if (bin>=t) {  // greater than _and equal_
      bin -= t;
      result += c;
    }
    t >>= 1; c >>= 1;
    if (!(i&3)) {
      update_char(result, index++);
      t = BCDbits[tenpower8--];
      c=8;
      result = '0';
    }
  }
  update_char('0' + bin, index);	/* last digit is easy */
}

/* tenpower8= 2(4 digits), 1(3 digits), 0(2 digit) see named constants in lcd.h */
void int2BCDn(uint16_t bin, lcdpos_t index, char tenpower8)
{
  uint8_t c;
  uint16_t result, t;

  while (tenpower8>=0) {  
      t = BCDbits[(uint8_t)tenpower8--];
      c=8; result = '0';	/* 1 digit of the number */

    if (bin>=t) {  // greater than _and equal_ 8..
      bin -= t; result += c;
    }
    t >>= 1; c >>= 1;

    if (bin>=t) {  // greater than _and equal_ 4..
      bin -= t; result += c;
    }
    t >>= 1; c >>= 1;

    if (bin>=t) {  // greater than _and equal_ 2..
      bin -= t; result += c;
    }
    t >>= 1; c >>= 1;

    if (bin>=t) {  // greater than _and equal_ 1..
      bin -= t; result += c;
    }

      update_char(result, index++);
  } // while

  if(bin > 9){
	update_char(OUT_OF_RANGE_CHAR, index);	/* out of range error */
  } else {
  	update_char('0' + bin, index);	/* last digit is easy */
  }
}

/* delays appr. p usec LAME busywait, should go to event-management */
void lcd_delay(uint16_t p)
{
  _delay_loop_2(4*p);

  //  uint16_t i;
  //  uint8_t j;
  //  for(i=0;i<p;i++)
  //    for (j=0;j<100;j++)	/* appr. usec TODO: much slower */
  //      asm("nop");
    /* should test sometimes for non-busy condition (but not in dummyloop) */ 
}

static char lcd_waitmode_flag;
/* after this is called with arg=1, we do not busywait after simple 
 * operations like lcd_putch() and lcd_goto() The mainloop likes this.
 * However from the more complex functions only int2BCD5() and the update_char..()
 * work with this (they use the lcd_cache)
 *
 * lcd_init() sets this to safe but inefficient default value: 0;
 */
void lcd_waitmode(char m)
{
	lcd_waitmode_flag = m;
}

__inline__ uint16_t lcd_wait_if(uint16_t q)
{
	return ( lcd_waitmode_flag ? 0 : q);
}

void toggle_E(void)
     //void toggle_E(void)
{
  // Enable pulsewidth must minimum be 230 ns, a few nops handles this.
  LCD_CTRL |= ENABLE;
  _delay_loop_1(2);
  LCD_CTRL &= ~ENABLE;
}

/* write command 7:4 bits assuming 8 bit interface */
void lcd_cmd8(uint8_t cmd, uint16_t wait)
{
  uint8_t t;
  if (wait) lcd_delay(wait);
    
  _delay_loop_1(10);
  LCD_CTRL &= ~( ENABLE | RS | DDIR );	/* 000 */
  t=LCD_DATA & 0x0f;	/* clear 7:4 bits */
  _delay_loop_1(10);
  LCD_DATA = t | (cmd & 0xf0);
  _delay_loop_1(10);
  toggle_E();	/* latch high bits */
  _delay_loop_1(10);
}

/* write command assuming 4 bit interface
TODO: should allow compile-time config 8 or 4 bits */
//__inline__ void lcd_cmd(uint8_t cmd, uint16_t wait)
void lcd_cmd(uint8_t cmd, uint16_t wait)
{
  uint8_t t;
  if (wait) 
    lcd_delay(wait);
  else
    busywait();

  _delay_loop_1(10);
  LCD_CTRL &= ~( ENABLE | RS | DDIR );	/* 000 */
  t=LCD_DATA & 0x0f;	/* clear 7:4 bits */
  _delay_loop_1(10);
  LCD_DATA = t | (cmd & 0xf0);
  _delay_loop_1(10);
  toggle_E();	/* latch high bits */
  // MIK TODO: read t again, and cli/sei it. An enable cycle must min be 500 ns anyway
  _delay_loop_1(10);
  LCD_DATA = t | (cmd << 4); /* TODO: if interrupt changes PORT value? */
  _delay_loop_1(10);
  toggle_E();	/* latch low bits */
  _delay_loop_1(10);
}

// __inline__ void lcd_data(uint8_t cmd, uint16_t wait)
void lcd_data(uint8_t cmd, uint16_t wait)
{
  uint8_t t;
  if (wait)
    lcd_delay(wait);
  else
    busywait();

  _delay_loop_1(10);
  LCD_CTRL &= ~( ENABLE | DDIR );    /* 0_0  NOTE that for data RS=1 */
  LCD_CTRL |= RS; /* we write data, not command */
  t=LCD_DATA & 0x0f;	/* clear 7:4 bits */
  _delay_loop_1(10);
  LCD_DATA = t | (cmd & 0xf0);
  _delay_loop_1(10);
  toggle_E();	/* latch high bits */
  _delay_loop_1(10);
  LCD_DATA = t | (cmd << 4); /* TODO: if interrupt changes PORT value? */
  _delay_loop_1(10);
  toggle_E();	/* latch low bits */
  _delay_loop_1(10);

  lcd_went_to ++; /* since cursorpos-autoincrease is turned on. TODO: check it. 
		     I assume this fails at the end of line, since LCD does NOT proceed to next line */
#if LCD_COLS == 16
  if ((lcd_went_to & (LCD_COLS - 1)) == 0) lcd_went_to = 255;
#elif LCD_COLS == 20
  if ((lcd_went_to % LCD_COLS) == 0) lcd_went_to = 255;
#else
#warning "Unknown LCD_COLS, the LCD won't work"
#endif

  /* should we put control lines back to LOW? 
     LCD_CTRL &= ~( ENABLE | RS | DDIR );    
  */
}

/*-----------------27.5.99 20:33--------------------
 *  Clear display
 *--------------------------------------------------*/
__inline__ void lcd_cls()
{
  lcd_cmd(0x01, 0);
  // TODO: this should clear the cache too

  //  lcd_cmd(0x01, 1000);
}

/*-----------------27.5.99 20:33--------------------
 *  Home position
 *--------------------------------------------------*/
void lcd_home()
{
  lcd_pos_to_peek = 0;
  lcd_cmd(0x02, 0);
  //  lcd_cmd(0x02, 1000);
  lcd_next_pos = 0;
}


/*-----------------27.5.99 20:33--------------------
 *  LCD & Cursor control (Blinking, ON/OFF,...)
 *--------------------------------------------------*/
void lcd_control(byte disonoff, byte curonoff, byte curblink)
{
  byte temp;
  temp=0x08;
  if (disonoff==1) temp|=0x04;
  if (curonoff==1) temp|=0x02;
  if (curblink==1) temp|=0x01;
  lcd_cmd(temp, 0);
  //  lcd_cmd(temp, 100);
}

/*-----------------27.5.99 20:31--------------------
 *  Goto position   (0x00 - 1.row, 1.col)
 *                  (0x40 - 2.row, 1.col)
 *--------------------------------------------------*/
void lcd_goto(byte mesto)
{
  lcd_cmd(0x80 | (mesto & 0x7f),  0);
  //  lcd_cmd(0x80 | (mesto & 0x7f),  lcd_wait_if(100));
}

/* wrapper around lcd_goto indexed as lcd_cache[] */
void lcd_go2index(lcdpos_t index)
{
  byte t;
#if LCD_COLS == 16 && LCD_LINES <=2

  t = index & (LCD_COLS - 1);
  if(index & LCD_COLS) t += LCD_2ND_LINE_ADDR;	/* LCD 2nd line */

#elif LCD_COLS == 20 && LCD_LINES == 4

  // the designer of the hd44780 4x20 char display propably had one of those monday mornings...
  // arghhhh, it is convoluted!

  // the 4x20 display works like a 2x40 display cut in half. Address 0 is the start 
  // of line 1. At the end of line 1, the cursor continues to line 3! The same is the
  // case for line 2 / line 4.

  if (index > (3*LCD_COLS - 1)) {
    t = index + 0x18; // magic constants, look in the datasheet for an explanation
  } else if (index > (2*LCD_COLS - 1)) {
    t = index - 0x14;
  } else if (index > (LCD_COLS - 1)) {
    t = index + 0x2c;
  } else {
    t = index;
  }
  // these address calculations makes it a lot easier to use the lcd,
  // now line 0 has address 0..19, line 1 has address 20..39 etc.

#endif

  lcd_goto(t);
  lcd_went_to = index;
}


__inline__ void lcd_setpos(lcdpos_t index)
{
  lcd_next_pos = index;
}

__inline__ lcdpos_t lcd_getpos(void)
{
  return lcd_next_pos;
}




// this is the old lcd_putch
__inline__ void lcd_write(byte data) {
  lcd_data(data, 0);
}


/*-----------------27.5.99 20:31--------------------
 *  Put character on LCD (in cache)
 *--------------------------------------------------*/
__inline__ void lcd_putch(byte data)
{ 
  update_char(data, lcd_next_pos);
}

/*-----------------27.5.99 20:30--------------------
 *  Display null terminated string
 *--------------------------------------------------*/
void lcd_putstr(byte *data)
{
  while(*data) {
    lcd_putch(*data);
    data++;
  }
}

// fetch string from flash memory and print to lcd
void lcd_putstr_P(const char *data)
{
  char c;
  // it might be necessary to use __elpm_inline()
  while( (c = PRG_RDB(data++)) ) {
    lcd_putch(c);
  }
}

/*-----------------27.5.99 20:29--------------------
 *  Display 8bit bin value
 *--------------------------------------------------*/
void printbin(byte x)
{
  byte i;
  for (i=128;i>0;i>>=1)
  {
    if ((x&i)==0) lcd_putch('0');
    else lcd_putch('1');
  }
}

/*-----------------1.6.99 16:39---------------------
 *  Display byte in hex
 *--------------------------------------------------*/
void printhex(byte i)
{
  byte hi,lo;
  
  hi = i & 0xF0;  // High nibble
  hi >>= 4;
  hi += '0';
  if (hi > '9')
    hi += 7;
  
  lo = (i & 0x0F) + '0';  // Low nibble
  if (lo > '9')
    lo += 7;
  
  lcd_putch(hi);
  lcd_putch(lo);
}

/*-----------------28.08.99 23:00-------------------
 *   Display 0-99
 * --------------------------------------------------*/
void print2(uint8_t x)
{
  uint8_t y;
   if (x < 100)
     {
       y = x/10;
       lcd_putch(y+0x30);
       x -= y*10;
       lcd_putch(x+0x30);
     }
   else 
     lcd_putstr("??");  // better to write a two character string
   //     lcd_putstr("Err");
}

/*-----------------27.5.99 20:29--------------------
 *  Display 0-999 unsigned char or int (8 or 16bit)
 *--------------------------------------------------*/
void print3(uint16_t x)
{
  uint16_t y;
  if (x<1000)
  {
    y=x/100;
    lcd_putch(y+0x30);

    x-=(y*100);
    y=x/10;
    lcd_putch(y+0x30);

    x-=(y*10);
    lcd_putch(x+0x30);
  }
  else lcd_putstr("Err");
}

/*-----------------27.5.99 20:30--------------------
 *  Display integer 0-65535 (16bit)
 *--------------------------------------------------*/
/*void print5(unsigned int x)
{
  unsigned int y;
    y=x/10000;lcd_putch(y+0x30);x-=(y*10000);
    y=x/1000;lcd_putch(y+0x30);x-=(y*1000);
    y=x/100;lcd_putch(y+0x30);x-=(y*100);
    y=x/10;lcd_putch(y+0x30);x-=(y*10);
    lcd_putch(x+0x30);
}*/

/*-----------------31.5.99 18:18--------------------
 *  Display unsigned long 0-4294967296 (32bit)
 *--------------------------------------------------*/
/*void print10(unsigned long x)
{
  unsigned int y;
    y=x/1000000000;lcd_putch(y+0x30);x-=(y*1000000000);
    y=x/100000000;lcd_putch(y+0x30);x-=(y*100000000);
    y=x/10000000;lcd_putch(y+0x30);x-=(y*10000000);
    y=x/1000000;lcd_putch(y+0x30);x-=(y*1000000);
    y=x/100000;lcd_putch(y+0x30);x-=(y*100000);
    y=x/10000;lcd_putch(y+0x30);x-=(y*10000);
    y=x/1000;lcd_putch(y+0x30);x-=(y*1000);
    y=x/100;lcd_putch(y+0x30);x-=(y*100);
    y=x/10;lcd_putch(y+0x30);x-=(y*10);
    lcd_putch(x+0x30);
}*/


__inline__ void lcd_ctrl_as_output(void)
{
	LCD_CTRLDDR |=  ENABLE | RS | DDIR;     /* outputs */
}

/* we drive the control lines LOW */
/*void  lcd_testLOW(void)
{
	lcd_ctrl_as_output();
	LCD_CTRL &= ~( ENABLE | RS | DDIR );
}*/

/* we drive the control lines HIGH */
/*void  lcd_testHIGH(void)
{
	lcd_ctrl_as_output();
	LCD_CTRL |= ( ENABLE | RS | DDIR );
}*/
/*void lcd_testrun(void)
{
	uint8_t t;
	for (t=0; t<64; t++){
		lcd_goto(t<<1);
		printhex(t);
	}
}*/

/*  Display initialization */
void lcd_init(void)
{
  uint8_t i;
  lcd_waitmode(0);
  lcd_pos_to_peek=0;
  lcd_next_pos=0;
  /* this dirties the whole LCD, and makes sure it will be refreshed: */
  //  update_chars("Hello This should clear the display hi hej", 0);
  //  for (i=0; i<LCD_CHARS; i++)
  //    lcd_putch(' ');

  for (i=0; i<LCD_CHARS; i++)
    lcd_cache[i] = ' '; // this will clear the cache

  for (i=0; i<(LCD_CHARS/8); i++)
    dirty[i] = 0xFF;  // set all dirty bits


  lcd_ctrl_as_output();
  LCD_CTRL &= ~( ENABLE | RS | DDIR );
  LCD_DATADDR |= 0xf0;

/*
  t=LCD_DATA & 0x0f;	
  LCD_DATA = t | 0x30;	// DL=8bit from now on 
  toggle_E();
*/

  // we might need some delay in case the lcd is slow starting
  for (i=0; i<60; i++)
    lcd_delay(2000);
  
  lcd_cmd8(0x30, 1000); // wait for 15 ms
  lcd_cmd8(0x30, 10000);  // wait for 4.1 ms
  lcd_cmd8(0x30, 500);   // wait for 100 us
  lcd_cmd8(0x20, 400);	/* now switched to 4 bit */
  lcd_cmd(0x28, 400);	/* 2 lines, 5x7 dots */
  //  lcd_cmd(0x0f, 40); 	/* display on, cursor on, cursor blink on = lcd_control(1,1,1) */
  lcd_cmd(0x0c, 400); 	/* display on, cursor off, cursor blink off = lcd_control(1,0,0) */
  lcd_cmd(0x06, 400);	/*  increment position, no display shift */

#if 0
  lcd_cmd8(0x30, 1000); // wait for 15 ms
  lcd_cmd8(0x30, 300);  // wait for 4.1 ms
  lcd_cmd8(0x30, 20);   // wait for 100 us
  lcd_cmd8(0x20, 20);	/* now switched to 4 bit */
  lcd_cmd(0x28, 20);	/* 2 lines, 5x7 dots */
  //  lcd_cmd(0x0f, 20); 	/* display on, cursor on, cursor blink on = lcd_control(1,1,1) */
  lcd_cmd(0x0c, 20); 	/* display on, cursor off, cursor blink off = lcd_control(1,0,0) */
  lcd_cmd(0x06, 20);	/*  increment position, no display shift */
#endif

  lcd_cls();
  lcd_went_to = 255;  // just let lcd_refresh take care of the cursor position
  lcd_setpos(0);
}

uint8_t lcd_busy(void) // read back busy flag (0x80)
{
  uint8_t t;
  LCD_DATADDR &= 0x0F; // data input
  LCD_CTRL &= ~( ENABLE | RS );
  LCD_CTRL |= DDIR;  /* we read back */
  _delay_loop_1(10);
  toggle_E();  /* is this necessary? */

  /* shell we need more delay ? */
  _delay_loop_1(10);
  t = 0x80 ;
  t &= LCD_DATAPIN ;  /* we just return the BUSY flag, not all 4 bits with 0xF0 */
  
  toggle_E();  // 4 bit mode; we need to toggle EN two times, else the lcd gets confused
  _delay_loop_1(10);
	
  LCD_CTRL &= ~( ENABLE | RS | DDIR );
  LCD_DATADDR |= 0xF0;  /* data out */
  return t;
}

/*-----------------28.08.99 22:49-------------------
 *   Simple printf function (no fp, and strings)
 *--------------------------------------------------*/
int printf(const char *format, ...){
  static const char hex[] = "0123456789ABCDEF";
  char format_flag;
  unsigned int u_val, div_val, base;
//   char *ptr;
  va_list ap;

  va_start (ap, format);
  for (;;) {
     while ((format_flag = *format++) != '%') {      // Until '%' or '\0'
       if (!format_flag) {
         va_end (ap);
         return (0);
       }
       lcd_putch (format_flag);
     }

     switch (format_flag = *format++) {
       case 'c': format_flag = va_arg(ap,int);
       default:  lcd_putch(format_flag); continue;

       case 'd': base = 10; div_val = 10000; goto CONVERSION_LOOP;
       case 'x': base = 16; div_val = 0x10;

       CONVERSION_LOOP:
       u_val = va_arg(ap,int);
       if (format_flag == 'd') {
         if (((int)u_val) < 0) {
           u_val = - u_val;
           lcd_putch ('-');
         }
         while (div_val > 1 && div_val > u_val) div_val /= 10;
       }
       do {
         lcd_putch (hex[u_val / div_val]);
         u_val %= div_val;
         div_val /= base;
       } while (div_val);
    }
  }
}
#endif
